#!/usr/bin/python
#
# debugger.py - Simple debugging client for Fuse
#
# Copyright (c) 2002 Anders Holmberg
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# Author contact information:
#
# Mail : aagholmberg@yahoo.se
#

# When requesting information Fuse responds with so called info blocks. An 
# info block contains register, memory and other data according to this
# specification. The info block is built of character data, e.g. the value
# of the register A is represented by two characters. 
#
# Offset Contents
#  0      A
#  2      F
#  4      B
#  6      C
#  8      D
#  10     E
#  12     H
#  14     L
#  16     IX
#  20     IY
#  24     SP
#  28     PC
#  32     I
#  34     R
#  36     IFF1
#  38     IFF2
#  40     IM
#  42     TSTATES
#  46     MEMORY DUMP ADDRESS
#  50     MEMORY DUMP OF 160 BYTES == 320 CHARACTERS
#

import socket, z80_disassembler

# Disassemble an info block and print 10 instructions to the screen
def disassemble( info_block ):

   # Inform the z80 disassembler of which info block that shall be examined
   z80_disassembler.set_data_block( info_block )
   
   # offset holds the index into the memory dump of the info block where the
   # next instructions is held (remember that each instructions can be 1-4
   # bytes long).
   offset = 0

   # Disassemble 10 instrunctions
   for i in range( 10 ):
      offset = offset + z80_disassembler.examine( offset )

# Build a character string holding the value of the registers
def build_register_string( data ):

   str = "\n-- REGISTERS --\n\n"

   str = str + " A  0x%c%c     F 0x%c%c\n"    % (data[0], data[1], data[2], data[3])   
   str = str + " B  0x%c%c     C 0x%c%c\n"    % (data[4], data[5], data[6], data[7])   
   str = str + " D  0x%c%c     E 0x%c%c\n"    % (data[8], data[9], data[10], data[11])   
   str = str + " H  0x%c%c     L 0x%c%c\n"    % (data[12], data[13], data[14], data[15])   
   str = str + " IX 0x%c%c%c%c\n"             % (data[16], data[17], data[18], data[19])   
   str = str + " IY 0x%c%c%c%c\n"             % (data[20], data[21], data[22], data[23])   
   str = str + " SP 0x%c%c%c%c\n"             % (data[24], data[25], data[26], data[27])   
   str = str + " PC 0x%c%c%c%c\n"             % (data[28], data[29], data[30], data[31])   
   str = str + " I  0x%c%c     R 0x%c%c\n"    % (data[32], data[33], data[34], data[35])   
   str = str + " IFF1 0x%c%c   IFF2 0x%c%c\n" % (data[36], data[37], data[38], data[39])   
   str = str + " IM 0x%c%c\n"                 % (data[40], data[41])   
   str = str + " TSTATES 0x%c%c%c%c\n"        % (data[42], data[43], data[44], data[45])   

   return str

# Build a character string holding the memory dump
def build_memory_string( data ):
   str = "\n-- MEMORY DUMP --\n\n" 

   address = int( data[46]+data[47]+data[48]+data[49], 16 )

   for row in range(10):
      row_string = " 0x%04x : " % (address + row*16)
      for column in range(16):
         row_string = row_string + "0x%c%c " % (data[50+(row*16+column)*2], data[50+1+(row*16+column)*2])
      
      str = str + row_string + "\n"
         
   return str

def print_help():
   print """

 -- FUSE DEBUGGER HELP --

 c      - continue
 h      - halt
 s      - single step

 b nnnn - set breakpoint to nnnn
 bd     - delete breakpoint

 d      - disassemble at PC
 d nnnn - disassemble at nnnn

 i      - print registers and memory dump at PC
 r      - print registers
 m nnnn - print memory dump at nnnn

 ---

   """


# Main program

if __name__ == "__main__":

   # Socket address tuple used for sending data
   send_addr = ( "127.0.0.1", 3491 )
   
   # Socket for reading and writing data (UDP)
   sockfd = socket.socket( socket.AF_INET, socket.SOCK_DGRAM )
   hostip = socket.gethostbyname( "127.0.0.1" )
   sockfd.bind( ( hostip, 3490 ) )  

   print "FUSE DEBUGGER (? for help)\n" 

   while 1:
      cmd = raw_input( " >> " )
  
      if len( cmd ) == 0:
         pass
 
      elif cmd[0] == '?':
         print_help()

      elif cmd[0] == 'h':
         print "HALT" 
         sockfd.sendto( "h", send_addr )

      elif cmd[0] == 'c':
         print "CONTINUE" 
         sockfd.sendto( "c", send_addr )

      elif cmd[0] == 'i':
         sockfd.sendto( "i", send_addr )
         data, address = sockfd.recvfrom( 400 )      
         print build_register_string( data )
         print build_memory_string( data )

      elif cmd[0] == 'm':
         sockfd.sendto( cmd, send_addr )
         data, address = sockfd.recvfrom( 400 )
         print build_memory_string( data )

      elif cmd[0] == 'b':
         sockfd.sendto( cmd, send_addr )

      elif cmd[0] == 's':
         sockfd.sendto( "s", send_addr )
         sockfd.sendto( "i", send_addr )
         data, address = sockfd.recvfrom( 400 )      
         print build_register_string( data )
         disassemble( data ) 

      elif cmd[0] == 'd':
         if len(cmd) == 1:
            sockfd.sendto( "i", send_addr )
         else:
            cmd = "m" + cmd[1:]
            sockfd.sendto( cmd, send_addr )
 
         data, address = sockfd.recvfrom( 400 )
         disassemble( data )

      elif cmd[0] == 'r':
         sockfd.sendto( "i", send_addr )
         data, address = sockfd.recvfrom( 400 )      
         print build_register_string( data )

      else:
         print "UNKNOWN CMD" 

